%%% BEGIN erlaunch.erl %%%
%%%
%%% erlaunch - execute selected Erlang functions from system command line
%%% Copyright (c)2003 Cat's Eye Technologies.  All rights reserved.
%%%
%%% Redistribution and use in source and binary forms, with or without
%%% modification, are permitted provided that the following conditions
%%% are met:
%%%
%%%   Redistributions of source code must retain the above copyright
%%%   notice, this list of conditions and the following disclaimer.
%%%
%%%   Redistributions in binary form must reproduce the above copyright
%%%   notice, this list of conditions and the following disclaimer in
%%%   the documentation and/or other materials provided with the
%%%   distribution.
%%%
%%%   Neither the name of Cat's Eye Technologies nor the names of its
%%%   contributors may be used to endorse or promote products derived
%%%   from this software without specific prior written permission.
%%%
%%% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
%%% CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
%%% INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
%%% MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
%%% DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
%%% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
%%% OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
%%% PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
%%% OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
%%% ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
%%% OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
%%% OUT OF THE USE OF THIS SOFTWARE, EVEN IF AsDVISED OF THE
%%% POSSIBILITY OF SUCH DAMAGE. 

%% @doc <code>erlaunch</code> is a system for efficiently executing
%% utility programs written for Erlang/OTP from the system command
%% line.
%%
%% <p>This module is the main Erlang component of the
%% <code>erlaunch</code> system.  It provides the long-running TCP/IP
%% server, which handles command requests, and a low-level I/O
%% interface for those commands.</p>
%%
%% @end

-module(erlaunch).
-define(ERLAUNCH_VERSION, '2004.0309').
-vsn('2004.0309').                % edoc chokes if I try to use the macro :(
-author('cpressey@catseye.mine.nu').
-copyright('Copyright (c)2003-2004 Cat`s Eye Technologies. All rights reserved.').

-export([server/1]).              % main server entry point, used by erlaunch.c
-export([fwrite/3, get_line/2]).  % raw i/o interface, used by 'raw' commands

%% @spec server([string()]) -> never_returns
%% @doc Main entry point for the long-running server component of
%% <code>erlaunch</code>.  This function is generally invoked by
%% <code>erlaunch.c</code> from <code>erl</code>'s command line, as in
%% <code>erl -run erlaunch server <i>version</i> <i>port</i> <i>cookie</i></code>.

server([Version, Port, Cookie]) ->
  case list_to_atom(Version) of
    ?ERLAUNCH_VERSION ->
      server0(Port, Cookie);
    _ ->
      init:stop()
  end.

server0(Port, Cookie) ->
  PortNo = list_to_integer(Port),
  {ok, CommandList} =
    file:consult(filename:join([code:priv_dir(?MODULE), "erlaunch.cmds"])),
  CommandDict = dict:from_list(CommandList),
  {ok, ListenSocket} = gen_tcp:listen(PortNo, [list,
    {reuseaddr, true}, {nodelay, true}, {active, false}, {packet, 4}]),
  loop(ListenSocket, Cookie, CommandDict).

loop(ListenSocket, Cookie, CommandDict) ->
  {ok, Socket} = gen_tcp:accept(ListenSocket),
  Pid = spawn(fun() -> serve(Socket, Cookie, CommandDict) end),
  gen_tcp:controlling_process(Socket, Pid),
  Pid ! go_ahead,
  loop(ListenSocket, Cookie, CommandDict).

serve(Socket, Cookie, CommandDict) ->
  receive
    go_ahead ->
      case catch serve0(Socket, Cookie, CommandDict) of
        {'EXIT', Reason} ->
          gen_tcp:send(Socket, [255, io_lib:format("~p", [Reason])]);
        _ ->
          ok
      end,
      gen_tcp:close(Socket)
  end.

serve0(Socket, Cookie, CommandDict) ->
  {ok, Cookie} = get_string(Socket),
  {ok, Command} = get_string(Socket),
  serve1(Socket, list_to_atom(Command), CommandDict).

serve1(Socket, stop, _CommandDict) ->
  gen_tcp:close(Socket),
  init:stop();
serve1(Socket, Command, CommandDict) ->
  case dict:find(Command, CommandDict) of
    {ok, CommandSpec} ->
      gen_tcp:send(Socket, [0]),
      serve2(Socket, CommandSpec);
    _ ->
      gen_tcp:send(Socket, [1]),
      not_found
  end.

serve2(Socket, {Module, Function, Style}) ->
  Args = read_list(Socket),
  {ok, Cwd} = get_string(Socket),
  EnvList = read_list(Socket),
  EnvDict = map_env(EnvList),
  case Style of
    raw ->
      Context = {Socket},
      case Module:Function(Context, Args, Cwd, EnvDict) of
        {exit, N} ->
          gen_tcp:send(Socket, [N + 1]),
          {exit, N}
      end;
    {mapped, ArgMap, ReturnMap} ->
      MappedArgs = map_args(ArgMap, Args),
      {ok, OWD} = file:get_cwd(),
      file:set_cwd(Cwd),
      OldEnv = os:getenv(),
      put_env(EnvList),
      Parent = self(),
      erlang:process_flag(trap_exit, true),
      Pid = spawn_link(fun() -> invoke(Parent, Module, Function, MappedArgs) end),
      Result = handle_io(Socket, Pid),
      ExitCode = find_match(ReturnMap, Result) + 1,
      put_env(OldEnv),
      ok = file:set_cwd(OWD),
      gen_tcp:send(Socket, [ExitCode]),
      ok
  end.

%%% I/O REDIRECTION %%%

invoke(Parent, Module, Function, Args) ->
  erlang:group_leader(Parent, self()),
  Result = apply(Module, Function, Args),
  Parent ! {self(), {done, Result}}.

handle_io(Socket, Pid) ->
  receive
    {Pid, {done, Result}} ->
      Result;
    {io_request, From, To, {put_chars, C}} ->
      gen_tcp:send(Socket, [0, C]),
      From ! {io_reply, To, ok},
      handle_io(Socket, Pid);
    {io_request, From, To, {put_chars, Module, Function, Args}} ->
      gen_tcp:send(Socket, [0, apply(Module, Function, Args)]),
      From ! {io_reply, To, ok},
      handle_io(Socket, Pid);
    {io_request, From, To, {get_line, Prompt}} ->
      gen_tcp:send(Socket, [0, atom_to_list(Prompt)]),
      case get_string(Socket) of
        {ok, String} ->
          From ! {io_reply, To, String};
        {error, eof} ->
          From ! {io_reply, To, eof};
        {error, Else} ->
          From ! {io_reply, To, {error, Else}}
      end,
      handle_io(Socket, Pid);
    {io_request, From, To, _IoRequest} ->
      From ! {io_reply, To, {error, not_supported}},
      handle_io(Socket, Pid);
    {'EXIT', _From, Reason} ->
      gen_tcp:send(Socket, [255, io_lib:format("~p", [Reason])]),
      {crashed, Reason}
  end.

%%% UTILITY FUNCTIONS %%%

map_args(Map, Args) ->
  map_args0(Map, list_to_tuple(Args), []).

map_args0([], _ArgTuple, Acc) ->
  lists:reverse(Acc);
map_args0([{arg, all} | Tail], ArgTuple, Acc) ->
  map_args0(Tail, ArgTuple, [tuple_to_list(ArgTuple) | Acc]);
map_args0([{arg, N} | Tail], ArgTuple, Acc) ->
  map_args0(Tail, ArgTuple, [element(N + 1, ArgTuple) | Acc]);
map_args0([{lit, N} | Tail], _ArgTuple, Acc) ->
  map_args0(Tail, _ArgTuple, [N | Acc]).

get_string(Socket) ->
  case gen_tcp:recv(Socket, 0) of
    {ok, [0 | String]} ->
      {ok, String};
    {ok, [EOF | _String]} when EOF > 0 ->
      {error, eof};
    {error, Reason} ->
      {error, Reason}
  end.

read_list(Socket) ->
  read_list(Socket, []).
read_list(Socket, Acc) ->
  case get_string(Socket) of
    {ok, ""} ->
      lists:reverse(Acc);
    {ok, Packet} ->
      read_list(Socket, [Packet | Acc])
  end.

map_env(List) ->
  List0 = lists:map(fun(String) ->
    split_env(String)
  end, List),
  dict:from_list(List0).

put_env(List) ->
  lists:foreach(fun(String) ->
    {Key, Value} = split_env(String),
    os:putenv(Key, Value)
  end, List).

split_env(String) ->
  case string:chr(String, $=) of
    0 ->
      {String, ""};
    N ->
      {string:substr(String, 1, N-1), string:substr(String, N+1)}
  end.

match('_', _) ->
  true;
match(W, X) when is_list(W), is_list(X) ->
  match_each(W, X);
match(W, X) when is_tuple(W), is_tuple(X) ->
  match_each(tuple_to_list(W), tuple_to_list(X));
match(X, X) ->
  true;
match(_, _) ->
  false.

match_each([], []) -> true;
match_each([WH | '_'], [XH | _]) ->
  match(WH, XH);
match_each([WH | WT], [XH | XT]) ->
  case match(WH, XH) of
    true ->
      match_each(WT, XT);
    false ->
      false
  end;
match_each(_, _) -> false.

find_match([{Wildcard, Result} | Tail], Term) ->
  case match(Wildcard, Term) of
    true ->
      Result;
    false ->
      find_match(Tail, Term)
  end.

%%% RAW INTERFACE %%%

%% @spec fwrite(erlaunch_context(), string(), [term()]) -> ok
%% @doc Writes formatted output to the caller in a 'raw' way.
%% Works much like <code>io:fwrite/3</code>, except takes an
%% <code>erlaunch_context()</code> instead of an <code>io()</code>.

fwrite({Socket}, Format, Args) ->
  gen_tcp:send(Socket, [0, io_lib:format(Format, Args)]),
  ok.

%% @spec get_line(erlaunch_context(), atom()) -> string() | eof
%% @doc Get unformatted input from the caller in a 'raw' way.
%% Works much like <code>io:get_line/2</code>, except takes an
%% <code>erlaunch_context()</code> instead of an <code>io()</code>.

get_line({Socket}, Prompt) ->
  gen_tcp:send(Socket, [0, atom_to_list(Prompt)]),
  case get_string(Socket) of
    {ok, String} ->
      String;
    {error, eof} ->
      eof;
    {error, Else} ->
      {error, Else}
  end.

%%% END of erlaunch.erl %%%
